@@ -19,30 +19,56 @@
 # MA 02110-1301, USA.
 
 package Lab;
+
 use strict;
+use warnings;
+use base qw(Exporter);
+
+# Lab format Version Number increased whenever incompatible changes
+# are done to the lab so that all packages are re-unpacked
+use constant LAB_FORMAT => 10;
+
+# Export now due to cicular depends between Lab and Lab::Package.
+our (@EXPORT, @EXPORT_OK, %EXPORT_TAGS);
+
+BEGIN {
+    @EXPORT = ();
+    %EXPORT_TAGS = (
+        constants => [qw(LAB_FORMAT)],
+        );
+    @EXPORT_OK = (
+        @{$EXPORT_TAGS{constants}}
+        );
+};
 
 use Util;
+# Only used by _populate_with_dist; remove when not needed
 use Lintian::Output qw(:messages);
 use Lintian::Command qw(spawn);
+use Lab::Package;
 
+use Cwd;
 use File::Temp;
 
+
+
 # Quiet "Name "main::LINTIAN_ROOT" used only once"
+# only used by _populate_with_dist
 () = ($main::LINTIAN_ROOT);
 
 my $LINTIAN_ROOT = $main::LINTIAN_ROOT;
 
 sub new {
-    my ( $class, $dir, $dist ) = @_;
+    my ( $class, $dir ) = @_;
 
     my $self = {};
     bless $self, $class;
 
-    $self->setup( $dir, $dist );
+    $self->_init( $dir );
     return $self;
 }
 
-
+# returns a truth value if the lab is initialized and exists
 sub is_lab {
     my ( $self ) = @_;
 
@@ -53,26 +79,35 @@ sub is_lab {
 	&& -d "$self->{dir}/info";
 }
 
-sub setup {
-    my ( $self, $dir, $dist ) = @_;
+sub _init {
+    my ( $self, $dir ) = @_;
 
     if ( $dir ) {
+        # Make sure we can always find it, even if we chdir around a lot.
+        my $absdir = Cwd::realpath($dir);
+        fail("Cannot determine the absolute path of $dir: $!") unless($absdir);
 	$self->{mode} = 'static';
-	$self->{dir} = $dir;
-	$self->{dist} = $dist;
-	
-	if (-d "$dir" && ! -d "$dir/changes") {
-	    mkdir("$dir/changes", 0777)
-		or fail("cannot create lab directory $dir/changes");
+	$self->{dir} = $absdir;
+
+        # This code is here fore BACKWARDS COMPATABILITY!
+        #  - we can kill it when LAB_FORMAT goes from 10 to 11.
+        #  Basically this auto-upgrades existing static labs to support changes files
+	if (-d "$absdir" && ! -d "$absdir/changes") {
+	    mkdir("$absdir/changes", 0777)
+		or fail("cannot create lab directory $absdir/changes");
 	}
     } else {
 	$self->{mode} = 'temporary';
 
 	my $created = 0;
 	for (1..10) {
-	    $dir = tmpnam();
+            my $absdir;
+            $dir = tmpnam(); # Not always absolute (e.g. if TMPDIR is relative)
+            $absdir = Cwd::realpath($dir);
+            fail("Cannot determine the absolute path of $dir: $!")
+                unless $absdir;
 
-	    if ($self->setup_force( $dir, $dist )) {
+	    if ($self->_do_setup( $dir )) {
 		$created = 1;
 		last;
 	    }
@@ -85,20 +120,21 @@ sub setup {
     return 1;
 }
 
+# Initialization method for static labs; must be called after new.
 sub setup_static {
     my ( $self ) = @_;
 
     unless ( $self->{mode} eq 'static' and $self->{dir} ) {
-	warning("no laboratory specified (need to define LINTIAN_LAB)");
+	warning('no laboratory specified (need to define LINTIAN_LAB)');
 	return 0;
     }
 
-    return $self->setup_force( $self->{dir}, $self->{dist} );
+    return $self->_do_setup( $self->{dir} );
 }
 
-
-sub setup_force {
-    my ( $self, $dir, $dist ) = @_;
+# backing sub for setup_static and (in some cases) _init
+sub _do_setup {
+    my ( $self, $dir ) = @_;
 
     return unless $dir;
 
@@ -131,67 +167,66 @@ sub setup_force {
 
     $self->{dir} = $dir;
     $ENV{'LINTIAN_LAB'} = $dir;
-    $self->populate_with_dist( $dist );
+    $self->_populate_with_dist();
 
     return 1;
 }
 
-sub populate_with_dist {
-    my ( $self, $dist ) = @_;
+# Deprecated; we need a better API for keeping the Lab in sync with a mirror.
+sub _populate_with_dist {
+    my ( $self ) = @_;
 
-    return 0 unless $dist;
+    return 0 unless $ENV{'LINTIAN_DIST'};
     return 0 unless $self->{dir};
 
-    debug_msg(2, "spawning list-binpkg, list-udebpkg and list-srcpkg since LINTIAN_DIST=$dist");
+    debug_msg(2, "spawning list-binpkg and list-srcpkg since LINTIAN_DIST=$ENV{'LINTIAN_DIST'}");
 
     my $v = $Lintian::Output::GLOBAL->verbose ? '-v' : '';
     my %opts = ( out => $Lintian::Output::GLOBAL->stdout );
     spawn(\%opts, ["$LINTIAN_ROOT/unpack/list-binpkg",
 		  "$self->{dir}/info/binary-packages", $v])
-	or fail("cannot create binary package list");
+	or fail('cannot create binary package list');
     spawn(\%opts, ["$LINTIAN_ROOT/unpack/list-srcpkg",
 		  "$self->{dir}/info/source-packages", $v])
-	or fail("cannot create source package list");
-    spawn(\%opts, ["$LINTIAN_ROOT/unpack/list-udebpkg",
-		  "$self->{dir}/info/udeb-packages", $v])
-	or fail("cannot create udeb package list");
+	or fail('cannot create source package list');
+    spawn(\%opts, ["$LINTIAN_ROOT/unpack/list-binpkg",
+		  "$self->{dir}/info/udeb-packages", '-u', $v])
+	or fail('cannot create udeb package list');
 
     return 1;
 }
 
+# Deletes the lab if (and only if) it exists and is a static lab
+# Returns a truth value on success
 sub delete_static {
     my ( $self ) = @_;
 
     unless ( $self->{mode} eq 'static' and $self->{dir} ) {
-	warning("no laboratory specified (need to define LINTIAN_LAB)");
+	warning('no laboratory specified (need to define LINTIAN_LAB)');
 	return 0;
     }
 
-    return $self->delete_force;
+    return $self->_do_delete;
 }
 
+# Deletes the lab if (and only if) it is a temporary lab
+# Returns a truth value on success (or it is not a temp lab)
 sub delete {
     my ( $self ) = @_;
 
     return 1 unless $self->{mode} eq 'temporary';
 
-    return $self->delete_force;
+    return $self->_do_delete;
 }
 
-# Remove is apparantly some reserved name...
-sub delete_force {
+# The backing sub for delete and delete_static
+sub _do_delete {
     my ( $self ) = @_;
 
     return 0 unless $self->{dir};
 
     v_msg("Removing $self->{dir} ...");
 
-    # since we will chdir in a moment, make the path of the lab absolute
-    unless ( $self->{dir} =~ m,^/, ) {
-	require Cwd;
-	$self->{dir} = Cwd::getcwd() . "/$self->{dir}";
-    }
-
     # chdir to root (otherwise, the shell will complain if we happen
     # to sit in the directory we want to delete :)
     chdir('/');
@@ -233,11 +268,59 @@ sub delete_force {
 		}
     }
 
-    $self->{dir} = "";
+    $self->{dir} = '';
 
     return 1;
 }
 
+
+{
+
+    # private helper variable.
+    my %pkg_types = (
+        'b' => 'binary',
+        'binary' => 'binary',
+        'c' => 'changes',
+        'changes' => 'changes',
+        's' => 'source',
+        'source' => 'source',
+        'u' => 'udeb',
+        'udeb' => 'udeb',
+    );
+
+    sub get_lab_package {
+        my ($self, $pkg_name, $pkg_version, $pkg_arch, $pkg_type, $pkg_path) = @_;
+        my $vpkg_type = $pkg_types{$pkg_type};
+        my $realpath = Cwd::realpath($pkg_path);
+        my $dir;
+        fail("Unknown package type $pkg_type") unless($vpkg_type);
+        fail("Could not resolve the path of $pkg_path") unless($realpath);
+        $dir = $self->{dir} . '/' . $vpkg_type . '/' . $pkg_name;
+        $dir .= '/' . $pkg_version if $self->_supports_multiple_versions();
+        $dir .= '/' . $pkg_arch if $self->_supports_multiple_architectures();
+        return new Lab::Package($self, $pkg_name, $pkg_version, $vpkg_type,
+                                $realpath, $dir);
+
+    }
+}
+
+# Returns a truth value if this is a "multi-version" Lab
+# This means that a new version of the same package can be extracted to the lab
+# without overwriting the old one.
+sub _supports_multiple_versions{
+    my ($self) = @_;
+    return $self->{mode} eq 'temporary';
+}
+
+# Returns a truth value if this is a "multi-arch" Lab
+# This means that (e.g.) an i386 and amd64 package with the same name will be stored
+# separatedly.  Otherwise unpacking a new package with different architecture will
+# override the old one.
+sub _supports_multiple_architectures{
+    my ($self) = @_;
+    return $self->{mode} eq 'temporary';
+}
+
 1;
 
 # vim: ts=4 sw=4 noet
