@@ -21,6 +21,7 @@
 
 package Util;
 use strict;
+use warnings;
 
 use Exporter;
 
@@ -43,7 +44,9 @@ BEGIN {
 	copy_dir
 	gunzip_file
 	touch_file
-	perm2oct);
+	perm2oct
+	check_path
+	resolve_pkg_path);
 }
 
 use FileHandle;
@@ -62,10 +65,22 @@ use Digest::MD5;
 #    (a hash contains one sections,
 #    keys in hash are lower case letters of control fields)
 sub parse_dpkg_control {
-    my ($CONTROL, $debconf_flag) = @_;
+    my @result;
+    my $c = sub { push @result, @_; };
+    _parse_dpkg_control_iterative( $c, @_);
+    return @result;
+}
+
+# parses a dpkg-control file like parse_dpkg_control, except
+# at the end of each section (or paragraph) it will pass the
+# section to a piece of code to handle it.  This allows reading
+# large dpkg-control based files without having the entire file
+# in memory.
+sub _parse_dpkg_control_iterative {
+    my ($code, $CONTROL, $debconf_flag) = @_;
 
-    my @data;
     my $cur_section = 0;
+    my $section = {};
     my $open_section = 0;
     my $last_tag;
 
@@ -78,8 +93,11 @@ sub parse_dpkg_control {
 	next if /^\#/;
 
 	# empty line?
-	if ((!$debconf_flag && m/^\s*$/) or ($debconf_flag && m/^$/)) {
+	if ((!$debconf_flag && m/^\s*$/) or ($debconf_flag && $_ eq '')) {
 	    if ($open_section) { # end of current section
+		# pass the current section to the handler
+		$code->($section);
+		$section = {};
 		$cur_section++;
 		$open_section = 0;
 	    }
@@ -101,7 +119,7 @@ sub parse_dpkg_control {
 	    $open_section = 1;
 
 	    my ($tag) = (lc $1);
-	    $data[$cur_section]->{$tag} = '';
+	    $section->{$tag} = '';
 
 	    $last_tag = $tag;
 	}
@@ -113,7 +131,7 @@ sub parse_dpkg_control {
 	    # immediately before or after the value and is ignored there.
 	    my ($tag,$value) = (lc $1,$2);
 	    $value =~ s/\s+$//;
-	    $data[$cur_section]->{$tag} = $value;
+	    $section->{$tag} = $value;
 
 	    $last_tag = $tag;
 	}
@@ -127,18 +145,24 @@ sub parse_dpkg_control {
 	    # field value are ignored.
 	    my $value = $1;
 	    $value =~ s/\s+$//;
-	    $data[$cur_section]->{$last_tag} .= "\n" . $value;
+	    $section->{$last_tag} .= "\n" . $value;
+	}
+	# None of the above => syntax error
+	else {
+	    my $message = "syntax error in section $cur_section";
+	    $message.= " after the tag $last_tag: $_" if defined $last_tag;
+	    fail($message);
 	}
     }
-
-    return @data;
+    # pass the last section (if not already done).
+    $code->($section) if $open_section;
 }
 
 sub read_dpkg_control {
     my ($file, $debconf_flag) = @_;
 
     if (not _ensure_file_is_sane($file)) {
-	return undef;
+	return;
     }
 
     open(my $CONTROL, '<', $file)
@@ -153,7 +177,7 @@ sub get_deb_info {
     my ($file) = @_;
 
     if (not _ensure_file_is_sane($file)) {
-	return undef;
+	return;
     }
 
     # dpkg-deb -f $file is very slow. Instead, we use ar and tar.
@@ -217,7 +241,7 @@ sub file_is_encoded_in_non_utf8 {
 	my ($file, $type, $pkg) = @_;
 	my $non_utf8 = 0;
 
-	open (ICONV, '-|', "env LANG=C iconv -f utf8 -t utf8 \Q$file\E 2>&1")
+	open (ICONV, '-|', "env LC_ALL=C iconv -f utf8 -t utf8 \Q$file\E 2>&1")
 	    or fail("failure while checking encoding of $file for $type package $pkg");
 	my $line = 1;
 	while (<ICONV>) {
@@ -305,7 +329,7 @@ sub fail {
     if (@_) {
 	$str = string('internal error', @_);
     } elsif ($!) {
-	$str = string('internal error', "$!");
+	$str = string('internal error', $!);
     } else {
 	$str = string('internal error');
     }
@@ -313,6 +337,95 @@ sub fail {
     die $str;
 }
 
+#check_path($command)>
+#
+#Return true if and only if $command is on the executable search path.
+#
+
+sub check_path {
+    my $command = shift;
+
+    return 0 unless exists $ENV{PATH};
+    for my $element (split ':', $ENV{PATH}) {
+	next unless length $element;
+	return 1 if -f "$element/$command" and -x _;
+    }
+    return 0;
+}
+
+#resolve_pkg_path($curdir, $dest)
+#
+# Using $curdir as current directory from the (package) root,
+# resolve $dest and return (the absolute) path to the destination.
+# Note that the result will never start with a slash, even if
+# $curdir or $dest does. Nor will it end with a slash.
+#
+# Note it will return '.' if the result is the package root.
+#
+# Returns a non-truth value, if it cannot safely resolve the path
+# (e.g. $dest would be outside the package root).
+#
+# Example:
+#  resolve_pkg_path('/usr/share/java', '../ant/file') eq  'usr/share/ant/file'
+#  resolve_pkg_path('/usr/share/java', '../../../usr/share/ant/file') eq  'usr/share/ant/file'
+#  resolve_pkg_path('/', 'usr/..') eq '.';
+#
+# The following will give a non-truth result:
+#  resolve_pkg_path('/usr/bin', '../../../../etc/passwd')
+#  resolve_pkg_path('/usr/bin', '/../etc/passwd')
+#
+sub resolve_pkg_path {
+    my ($curdir, $dest) = @_;
+    my (@cc, @dc);
+    my $target;
+    $dest =~ s,//++,/,o;
+    # short curcuit $dest eq '/' case.
+    return '.' if $dest eq '/';
+    # remove any initial ./ and trailing slashes.
+    $dest =~ s,^\./,,o;
+    $dest =~ s,/$,,o;
+    if ($dest =~ m,^/,o){
+	# absolute path, strip leading slashes and resolve
+	# as relative to the root.
+	$dest =~ s,^/,,o;
+	return resolve_pkg_path('/', $dest);
+    }
+
+    # clean up $curdir (as well)
+    $curdir =~ s,//++,/,o;
+    $curdir =~ s,/$,,o;
+    $curdir =~ s,^/,,o;
+    $curdir =~ s,^\./,,o;
+    # Short circuit the '.' (or './' -> '') case.
+    if ($dest eq '.' or $dest eq '') {
+	$curdir =~ s,^/,,o;
+	return '.' unless $curdir;
+	return $curdir;
+    }
+    # Relative path from src
+    @dc = split(m,/,o, $dest);
+    @cc = split(m,/,o, $curdir);
+    # Loop through @dc and modify @cc so that in the
+    # end of the loop, @cc will contain the path that
+    # - note that @cc will be empty if we end in the
+    # root (e.g. '/' + 'usr' + '..' -> '/'), this is
+    # fine.
+    while ($target = shift @dc) {
+	if($target eq '..') {
+	    # are we out of bounds?
+	    return '' unless @cc;
+	    # usr/share/java + '..' -> usr/share
+	    pop @cc;
+	} else {
+	    # usr/share + java -> usr/share/java
+	    push @cc, $target;
+	}
+    }
+    return '.' unless @cc;
+    return join '/', @cc;
+}
+
+
 1;
 
 # Local Variables:
